import axinite as ax
import numpy as np
from numba import jit

def load(delta, limit, backend, *bodies, t=0, modifier=None, action=None, action_frequency=200):
    """
    Loads a simulation from a backend.

    Args:
        delta (np.float64): The change in time between each step.
        limit (np.float64): The limit of the simulation.
        backend (function): The backend to load the simulation from.
        t (np.float64, optional): The initial time. Defaults to 0.0.
        modifier (function, optional): The modifier to apply to forces in the simulation. Defaults to None.
        action (function, optional): An action to call with frequency `action_frequency`. Defaults to None.
        action_frequency (int, optional): The frequency at which to call the `action`. Defaults to 200.

    Returns:
        list[ax.Body]: A list of Body objects after the simulation.
    """
    try:
        _bodies = backend(delta, limit, ax.get_inner_bodies(bodies), action=action, modifier=modifier, t=t, action_frequency=action_frequency)
    except TypeError as e:
        print("Error: Invalid backend.")
        raise e
    __bodies = []
    for body in _bodies:
        _body = ax.Body(str(body["n"]), body["m"], limit, delta)
        _body._inner = body
        __bodies.append(_body)
    return __bodies



import numpy as np
import axinite as ax
from numba import njit
import ast, operator

G = 6.67430e-11

def body_dtype(limit: np.float64, delta: np.float64) -> np.dtype:
    """
    Returns the data type for a body.

    Args:
        limit (np.float64): The length of the simulation in seconds.
        delta (np.float64): The frequency at which the simulation should be computed in seconds.

    Returns:
        np.dtype: The data type for a body.
    """
    try: 
        return np.dtype([
            ("n", "U20"),
            ("m", np.float64),
            ("r", np.float64, (int(limit/delta), 3)),
            ("v", np.float64, (int(limit/delta), 3))
        ])
    except ValueError as e:
        print("Error: limit/delta must be an integer.")
        raise e

def _body(limit: np.float64, delta: np.float64, name: str, mass: np.float64) -> np.ndarray: 
    """
    Creates a new body.

    Args:
        limit (np.float64): The length of the simulation in seconds.
        delta (np.float64): The frequency at which the simulation should be computed in seconds.
        name (str): The name of the body.
        mass (np.float64): The mass of the body in kilograms.

    Returns:
        np.ndarray: The new body.
    """
    try: return np.array((name, mass, np.zeros((int(limit/delta), 3)), np.zeros((int(limit/delta), 3))), dtype=ax.body_dtype(limit, delta))
    except ValueError as e: 
        print("Error: limit/delta must be an integer.")
        raise e

def get_inner_bodies(bodies: list[ax.Body]) -> np.ndarray:
    """
    Returns the inner representations of a list of bodies.

    Args:
        bodies (list[ax.Body]): A list of Body objects.

    Returns:
        np.ndarray: The inner representation of the bodies.
    """
    _bodies = ()
    for body in bodies: _bodies += (body._inner,)
    return np.array(_bodies)

def create_outer_bodies(bodies: np.ndarray, limit: np.float64, delta: np.float64) -> list[ax.Body]:
    """
    Creates outer body representations from inner bodies.

    Args:
        bodies (np.ndarray): An array of inner body representations.
        limit (np.float64): The length of the simulation in seconds.
        delta (np.float64): The frequency at which the simulation should be computed in seconds.

    Returns:
        list[ax.Body]: A list of Body objects.
    """
    _bodies = []
    for body in bodies:
        _body = ax.Body(str(body["n"]), body["m"], limit, delta)
        _body._inner = body
        _bodies.append(_body)
    return _bodies

@njit 
def vector_magnitude_jit(vec: np.ndarray) -> np.float64:
    """
    Calculates the magnitude of a vector.

    Args:
        vec (np.ndarray): The vector to calculate the magnitude of.

    Returns:
        np.float64: The magnitude of the vector.
    """
    return np.sqrt(np.sum(vec**2))

@njit
def unit_vector_jit(vec: np.ndarray) -> np.ndarray:
    """Calculates the unit vector of a vector.

    Args:
        vec (np.ndarray): The vector to calculate the unit vector of.

    Returns:
        np.ndarray: The unit vector of the vector.
    """
    mag = vector_magnitude_jit(vec)
    return vec / mag if mag > 0 else vec

@njit
def gravitational_force_jit(m1: np.float64, m2: np.float64, r: np.ndarray, epsilon: np.float64 = 1e-10) -> np.ndarray:
    """Calculates the gravitational force between two bodies.

    Args:
        m1 (u.Quantity): The mass of the first body.
        m2 (u.Quantity): The mass of the second body.
        r (np.ndarray): The vector between the two bodies.
        epsilon (np.float64, optional): The softening parameter. Defaults to 1e-10.

    Returns:
        np.ndarray: The gravitational force between the two bodies.
    """
    mag = vector_magnitude_jit(r)
    if mag < epsilon:
        return np.zeros(3)
    return -G *((m1 * m2) / mag**2) * unit_vector_jit(r)

def timestep(t: np.float64, delta: np.float64) -> int:
    """
    Computes the time step.

    Args:
        t (np.float64): The current time.
        delta (np.float64): The frequency at which the simulation should be computed in seconds.

    Returns:
        int: The new time.
    """
    return int(t / delta)

def timesteps(limit: np.float64, delta: np.float64) -> int:
    """
    Computes the number of time steps.

    Args:
        limit (np.float64): The length of the simulation in seconds.
        delta (np.float64): The frequency at which the simulation should be computed in seconds.

    Returns:
        int: The number of time steps.
    """
    return int(limit / delta)

def safe_eval(expr):
    """Safely evaluate a basic math expression."""
    allowed_operators = {
        ast.Add: operator.add,
        ast.Sub: operator.sub,
        ast.Mult: operator.mul,
        ast.Div: operator.truediv,
        ast.Pow: operator.pow,
        ast.BitXor: operator.xor,
        ast.USub: operator.neg,
    }

    def eval_node(node):
        if isinstance(node, ast.Num):  # <number>
            return node.n
        elif isinstance(node, ast.BinOp):  # <left> <operator> <right>
            return allowed_operators[type(node.op)](eval_node(node.left), eval_node(node.right))
        elif isinstance(node, ast.UnaryOp):  # <operator> <operand> e.g., -1
            return allowed_operators[type(node.op)](eval_node(node.operand))
        else:
            raise TypeError(node)

    node = ast.parse(expr, mode='eval').body
    return eval_node(node)

def interpret_time(string: str) -> np.float64:
    """Interprets a string as a time in seconds.

    Args:
        string (str): The string to interpret.

    Returns:
        float: The time in seconds.
    """
    if type(string) is float or type(string) is int: return string
    try:
        return safe_eval(string)
    except:
        units = {"s": 1, "min": 60, "hr": 3600, "d": 86400, "yr": 31536000}
        total_time = 0
        for part in string.split('+'):
            part = part.strip()
            for unit, factor in units.items():
                if part.endswith(unit):
                    part = part.removesuffix(unit)
                    total_time += float(part) * factor
                    break
            else:
                try:
                    total_time += float(part)
                except ValueError as e:
                    print(f"Unrecognized time unit in '{part}'")
                    raise e
        return total_time

def interpret_mass(string: str) -> np.float64:
    """Interprets a string as a mass in kilograms.

    Args:
        string (str): The string to interpret.

    Returns:
        float: The mass in kilograms.
    """
    if type(string) is float or type(string) is int: return string
    try:
        return safe_eval(string)
    except:
        units = {"kg": 1, "g": 1/1000, "t": 1000}
        total_mass = 0
        for part in string.split('+'):
            part = part.strip()
            for unit, factor in units.items():
                if part.endswith(unit):
                    part = part.removesuffix(unit)
                    total_mass += float(part) * factor
                    break
            else:
                try:
                    total_mass += float(part)
                except ValueError as e:
                    print(f"Unrecognized mass unit in '{part}'")
                    raise e
        return total_mass

def interpret_distance(string: str) -> np.float64:
    """Interprets a string as a distance in meters.

    Args:
        string (str): The string to interpret.

    Returns:
        float: The distance in meters.
    """
    if type(string) is float or type(string) is int: return string
    try:
        return safe_eval(string)
    except:
        units = {"m": 1, "km": 1000, "cm": 1/100, "mm": 1/1000, "μm": 1/1000000, "nm": 1/1000000000}
        total_distance = 0
        for part in string.split('+'):
            part = part.strip()
            for unit, factor in units.items():
                if part.endswith(unit):
                    part = part.removesuffix(unit)
                    total_distance += float(part) * factor
                    break
            else:
                try:
                    total_distance += float(part)
                except ValueError as e:
                    print(f"Unrecognized distance unit in '{part}'")
                    raise e
        return total_distance

def time_to(time: np.float64, unit: str, round_digits: int=-1) -> str:
    """Converts a time to a specific unit.

    Args:
        time (np.float64): The time to convert.
        unit (str): The unit to convert to.
        round_digits (int, optional): The number of digits to round to. Defaults to -1.

    Returns:
        str: The converted time.
    """
    if unit == "min":
        converted_time = time / 60
    elif unit == "hr":
        converted_time = time / 3600
    elif unit == "d":
        converted_time = time / 86400
    elif unit == "yr":
        converted_time = time / 31536000
    else:
        converted_time = time

    if round_digits >= 0:
        converted_time = round(converted_time, round_digits)

    return f"{converted_time}{unit}"

def mass_to(mass: np.float64, unit: str, round_digits: int=-1) -> str:
    """Converts a mass to a specific unit.

    Args:
        mass (np.float64): The mass to convert.
        unit (str): The unit to convert to.
        round_digits (int, optional): The number of digits to round to. Defaults to -1.

    Returns:
        str: The converted mass.
    """
    if unit == "g":
        converted_mass = mass * 1000
    elif unit == "t":
        converted_mass = mass / 1000
    else:
        converted_mass = mass

    if round_digits >= 0:
        converted_mass = round(converted_mass, round_digits)

    return f"{converted_mass}{unit}"

def distance_to(distance: np.float64, unit: str, round_digits: int=-1) -> str:
    """Converts a distance to a specific unit.

    Args:
        distance (np.float64): The distance to convert.
        unit (str): The unit to convert to.
        round_digits (int, optional): The number of digits to round to. Defaults to -1.

    Returns:
        str: The converted distance.
    """
    if unit == "km":
        converted_distance = distance / 1000
    elif unit == "cm":
        converted_distance = distance * 100
    elif unit == "mm":
        converted_distance = distance * 1000
    elif unit == "μm":
        converted_distance = distance * 1000000
    elif unit == "nm":
        converted_distance = distance * 1000000000
    else:
        converted_distance = distance

    if round_digits >= 0:
        converted_distance = round(converted_distance, round_digits)

    return f"{converted_distance}{unit}"

@njit
def clip_scalar(scalar, min, max):
    """
    Clips a scalar value between a minimum and maximum value.

    Args:
        scalar (float): The scalar value to clip.
        min (float): The minimum value.
        max (float): The maximum value.

    Returns:
        float: The clipped scalar value.
    """
    if scalar < min: return min
    if scalar > max: return max
    return scalar

def state(body, n):
    return body._inner["r"][n], body._inner["v"][n]

def round_limit(limit, delta):
    return round(limit / delta) * delta

def round_time(t, delta):
    return round(t / delta) * delta

@njit
def gravitational_forces(bodies, body, i, n):
    f = np.zeros(3)
    for j, other in enumerate(bodies):
        if i != j: f += ax.gravitational_force_jit(body["m"], other["m"], body["r"][n] - other["r"][n])
    return f

@njit 
def locate_body(body, bodies):
    for i, b in enumerate(bodies):
        if b["n"] == body["n"]:
            return i
    return -1


import axinite.tools as axtools
import json

def save(args: axtools.AxiniteArgs, path: str):
    """
    Saves the simulation data to a file.

    Args:
        args (axtools.AxiniteArgs): The arguments for the simulation.
        path (str): The path to save the simulation data to.
    """

    with open(path, "w") as file:
        file.write(saves(args))

def saves(args: axtools.AxiniteArgs):
    """
    Converts the simulation data to a JSON string.

    Args:
        args (axtools.AxiniteArgs): The arguments for the simulation.

    Returns:
        str: The JSON string representation of the simulation data.
    """

    data = {
        "name": args.name,
        "delta": float(args.delta),
        "limit": float(args.limit),
        "t": float(args.t),
        "bodies": []
    }

    for body in args.bodies: 
        body_data = {
            "name": str(body.name),
            "mass": float(body.mass),
            "radius": float(body.radius),
            "r": [],
            "v": []
        }

        for i, r in enumerate(body._inner["r"]):
            v = body._inner["v"][i]
            _r = [float(_v) for _v in r]
            body_data["r"].append(_r)
            _v = [float(__v) for __v in v]
            body_data["v"].append(_v)

        if body.color != None:
            body_data["color"] = body.color
        if body.retain != None:
            body_data["retain"] = int(body.retain)
        if body.light != None:
            body_data["light"] = body.light

        data["bodies"].append(body_data)

    if args.radius_multiplier is not None:
        data["radius_multiplier"] = int(args.radius_multiplier)

    if args.rate is not None:
        data["rate"] = args.rate

    if args.retain is not None:
        data["retain"] = args.retain
    
    if args.frontend_args != {}:
        data["frontend_args"] = args.frontend_args

    return json.dumps(data)


import axinite.tools as axtools
import axinite as ax
import matplotlib.pyplot as plt
import os, signal
from itertools import cycle

color_cycle = cycle(['r', 'b', 'g', 'orange', 'purple', 'yellow', 'gray', 'black'])

def mpl_frontend(args: axtools.AxiniteArgs, mode: str, type: str = "2D", **kwargs):
    if type == "2D" and mode == "show":
        return mpl_2d_static(args, **kwargs)
    elif type == "2D" and mode == "live":
        return mpl_2d_live(args, **kwargs)
    elif type == "3D" and mode == "show":
        return mpl_3d_static(args, **kwargs)
    else:
        raise Exception("Unsupported mode or type for mpl_frontend")

def mpl_2d_static(args: axtools.AxiniteArgs, **kwargs):
    if args.rate is None:
        args.rate = 100
    if args.radius_multiplier is None:
        args.radius_multiplier = 1
    if args.retain is None:
       args.retain = 200

    fig, axs = plt.subplots(2, 2, figsize=(15, 15))
    plt.ion()
    fig.show()

    for _ax in axs.flat:
        _ax.set_aspect('equal', adjustable='datalim')

    def update_plot(body: axtools.Body):
        color = axtools.string_to_color(body.color, "mpl") if body.color != "" else next(color_cycle)
        trajectory_x, trajectory_y, trajectory_z = ([], [], [])
        for r in body._inner["r"]:
            trajectory_x.append(r[0])
            trajectory_y.append(r[1])
            trajectory_z.append(r[2])

        axs[0, 0].plot(trajectory_x, trajectory_y, label=body.name, color=color)
        axs[0, 1].plot(trajectory_x, trajectory_z, label=body.name, color=color)
        axs[1, 0].plot(trajectory_y, trajectory_z, label=body.name, color=color)

        if trajectory_x and trajectory_y and trajectory_z:
            radius = body.radius * body.radius_multiplier * args.radius_multiplier
            start_circle_xy = plt.Circle((trajectory_x[0], trajectory_y[0]), radius, color=color, fill=True, alpha=0.3)
            start_circle_xz = plt.Circle((trajectory_x[0], trajectory_z[0]), radius, color=color, fill=True, alpha=0.3)
            start_circle_yz = plt.Circle((trajectory_y[0], trajectory_z[0]), radius, color=color, fill=True, alpha=0.3)
            axs[0, 0].add_artist(start_circle_xy)
            axs[0, 1].add_artist(start_circle_xz)
            axs[1, 0].add_artist(start_circle_yz)

            axs[0, 0].plot(trajectory_x[-1], trajectory_y[-1], 'x', color=color)
            axs[0, 1].plot(trajectory_x[-1], trajectory_z[-1], 'x', color=color)
            axs[1, 0].plot(trajectory_y[-1], trajectory_z[-1], 'x', color=color)

        axs[0, 0].set_xlabel('X')
        axs[0, 0].set_ylabel('Y')
        axs[0, 1].set_xlabel('X')
        axs[0, 1].set_ylabel('Z')
        axs[1, 0].set_xlabel('Y')
        axs[1, 0].set_ylabel('Z')

        axs[0, 0].legend()
        axs[0, 1].legend()
        axs[1, 0].legend()
        plt.tight_layout()

        plt.draw()
        plt.pause(0.001)

    def show():
        axs[0, 0].set_title('XY Plane')
        axs[0, 1].set_title('XZ Plane')
        axs[1, 0].set_title('YZ Plane')

        plt.show(block=True)

    def stop():
        os.kill(os.getpid(), signal.SIGINT)

    return update_plot, show, stop

def mpl_2d_live(args: axtools.AxiniteArgs, **kwargs):
    if args.rate is None:
        args.rate = 100
    if args.radius_multiplier is None:
        args.radius_multiplier = 1
    if args.retain is None:
       args.retain = 200

    fig, axs = plt.subplots(2, 2, figsize=(15, 15))
    plt.ion()
    fig.show()

    for _ax in axs.flat:
        _ax.set_aspect('equal', adjustable='datalim')

    radii = {body.name: body.radius * body.radius_multiplier * args.radius_multiplier for body in args.bodies}
    colors = {body.name: axtools.string_to_color(body.color, "mpl") if body.color != "" else next(color_cycle) for body in args.bodies}

    def setup_plot():
        axs[0, 0].set_title('XY Plane')
        axs[0, 1].set_title('XZ Plane')
        axs[1, 0].set_title('YZ Plane')

        axs[0, 0].set_xlabel('X')
        axs[0, 0].set_ylabel('Y')
        axs[0, 1].set_xlabel('X')
        axs[0, 1].set_ylabel('Z')
        axs[1, 0].set_xlabel('Y')
        axs[1, 0].set_ylabel('Z')

        plt.tight_layout()

    def update_plot(bodies, t, **kwargs):
        try:
            axs[0, 0].clear()
            axs[0, 1].clear()
            axs[1, 0].clear()
            axs[1, 1].clear()

            bodies = ax.create_outer_bodies(bodies, kwargs["limit"], kwargs["delta"])
            n = kwargs["n"]

            for body in bodies:
                trajectory_x, trajectory_y, trajectory_z = ([], [], [])
                for r in body.rs[:kwargs["n"]]:
                    trajectory_x.append(r[0])
                    trajectory_y.append(r[1])
                    trajectory_z.append(r[2])

                axs[0, 0].plot(trajectory_x, trajectory_y, label=body.name, color=colors[body.name])
                axs[0, 1].plot(trajectory_x, trajectory_z, label=body.name, color=colors[body.name])
                axs[1, 0].plot(trajectory_y, trajectory_z, label=body.name, color=colors[body.name])

                if trajectory_x and trajectory_y and trajectory_z:
                    circle_xy = plt.Circle((trajectory_x[-1], trajectory_y[-1]), radii[body.name], color=colors[body.name], fill=True, alpha=0.5)
                    circle_xz = plt.Circle((trajectory_x[-1], trajectory_z[-1]), radii[body.name], color=colors[body.name], fill=True, alpha=0.5)
                    circle_yz = plt.Circle((trajectory_y[-1], trajectory_z[-1]), radii[body.name], color=colors[body.name], fill=True, alpha=0.5)
                    axs[0, 0].add_artist(circle_xy)
                    axs[0, 1].add_artist(circle_xz)
                    axs[1, 0].add_artist(circle_yz)
            axs[0, 0].legend()
            axs[0, 1].legend()
            axs[1, 0].legend()

            setup_plot()

            plt.draw()
            plt.pause(1 / args.rate)
        except KeyboardInterrupt:
            os.kill(os.getpid(), signal.SIGINT)

    def show():
        setup_plot()
        plt.show(block=True)

    def stop():
        os.kill(os.getpid(), signal.SIGINT)

    return update_plot, show, stop

def mpl_3d_static(args: axtools.AxiniteArgs, **kwargs):
    if args.rate is None:
        args.rate = 100
    if args.radius_multiplier is None:
        args.radius_multiplier = 1
    if args.retain is None:
       args.retain = 200

    fig = plt.figure(figsize=(15, 15))
    ax = fig.add_subplot(111, projection='3d')
    plt.ion()
    fig.show()

    def update_plot(body: axtools.Body):
        color = axtools.string_to_color(body.color, "mpl") if body.color != "" else next(color_cycle)
        trajectory_x, trajectory_y, trajectory_z = ([], [], [])
        for r in body._inner["r"]:
            trajectory_x.append(r[0])
            trajectory_y.append(r[1])
            trajectory_z.append(r[2])

        ax.plot(trajectory_x, trajectory_y, trajectory_z, label=body.name, color=color)

        if trajectory_x and trajectory_y and trajectory_z:
            ax.scatter(trajectory_x[0], trajectory_y[0], trajectory_z[0], color=color, alpha=0.6)
            ax.scatter(trajectory_x[-1], trajectory_y[-1], trajectory_z[-1], color=color, marker='x')

        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        ax.set_zlabel('Z')

        ax.legend()
        fig.tight_layout(rect=[0, 0, 1, 0.95])

        plt.draw()
        plt.pause(0.001)

    def show():
        ax.set_title('3D Trajectory')
        plt.show(block=True)

    def stop():
        os.kill(os.getpid(), signal.SIGINT)

    return update_plot, show, stop



import axinite.tools as axtools
import axinite as ax
import numpy as np
from numba import njit

def intercept(a: axtools.Body, b: axtools.Body, speed_range: tuple[np.float64, np.float64], n_timesteps, delta, verbose=False):
    """
    Attempt to intercept body 'a' with body 'b' within a given speed range and time steps.

    Args:
        a (axtools.Body): The intercepting body.
        b (axtools.Body): The target body.
        speed_range (tuple[np.float64, np.float64]): The range of speeds to consider for interception.
        n_timesteps (int): The number of time steps to consider.
        delta (float): The time step interval.
        verbose (bool): Whether to print verbose output.

    Returns:
        tuple or None: Interception details (speed, position, timestep, unit vector) if found, otherwise None.
    """
    if verbose: print(f"Attempting to intercept {a.name} with {b.name}")
    dtype = np.dtype([
        ("n", "U20"),
        ("m", np.float64),
        ("r", np.float64, (n_timesteps, 3)),
        ("v", np.float64, (n_timesteps, 3)),
        ("rad", np.float64)
    ])
    return _intercept(
        np.array((a.name, a.mass, a._inner["r"], a._inner["v"], a.radius), dtype=dtype),
        np.array((b.name, b.mass, b._inner["r"], b._inner["v"], b.radius), dtype=dtype),
        speed_range, delta, n_timesteps, verbose
    )

@njit
def _intercept(a, b, speed_range, delta, n, verbose):
    """
    Numba-compiled function to attempt to intercept body 'a' with body 'b' within a given speed range and time steps.

    Args:
        a (np.ndarray): The intercepting body.
        b (np.ndarray): The target body.
        speed_range (tuple[np.float64, np.float64]): The range of speeds to consider for interception.
        delta (float): The time step interval.
        n (int): The number of time steps to consider.
        verbose (bool): Whether to print verbose output.

    Returns:
        tuple or None: Interception details (speed, position, timestep, unit vector) if found, otherwise None.
    """
    _n = float(n)
    for i, r in enumerate(a["r"]):
        if verbose: print(f"Checking timestep {i} ({int(i/_n*100)}%)")
        unit_vector = ax.unit_vector_jit(r - b["r"][0])
        magnitude = ax.vector_magnitude_jit(r - b["r"][0])
        speed = magnitude / ((i + 1) * delta)
        if verbose: print(f"Speed: {int(speed)}\033[F\033[F")
        if speed_range[0] <= speed <= speed_range[1]: 
            print(f"\n\nFound!")
            return (speed, r, i, unit_vector)
    print("\n\nNot found")

    return None

def intercept_at(n, a, b, delta, speed_range):
    """
    Attempt to intercept body 'a' with body 'b' at a specific timestep.

    Args:
        n (int): The specific timestep to consider.
        a (axtools.Body): The intercepting body.
        b (axtools.Body): The target body.
        delta (float): The time step interval.
        speed_range (tuple[np.float64, np.float64]): The range of speeds to consider for interception.

    Returns:
        tuple or None: Interception details (speed, position, timestep, unit vector) if found, otherwise None.
    """
    return _intercept_at(n, a._inner, b._inner, delta, speed_range)

@njit
def _intercept_at(n, a, b, delta, speed_range):
    """
    Numba-compiled function to attempt to intercept body 'a' with body 'b' at a specific timestep.

    Args:
        n (int): The specific timestep to consider.
        a (np.ndarray): The intercepting body.
        b (np.ndarray): The target body.
        delta (float): The time step interval.
        speed_range (tuple[np.float64, np.float64]): The range of speeds to consider for interception.

    Returns:
        tuple or None: Interception details (speed, position, timestep, unit vector) if found, otherwise None.
    """
    unit_vector = ax.unit_vector_jit(b["r"][0] - a["r"][n])
    magnitude = ax.vector_magnitude_jit(b["r"][0] - a["r"][n])
    speed = magnitude / ((n + 1) * delta)
    if speed_range[0] <= speed <= speed_range[1]:
        return (speed, a["r"][n], n, unit_vector)
    return None


import axinite as ax
import axinite.analysis as axana
import numpy as np
from numba import jit

@jit
def quaternion_multiply(q1, q2):
    """
    Multiply two quaternions.

    Args:
        q1 (np.ndarray): The first quaternion.
        q2 (np.ndarray): The second quaternion.

    Returns:
        np.ndarray: The product of the two quaternions.
    """
    w1, x1, y1, z1 = q1
    w2, x2, y2, z2 = q2
    w = w1 * w2 - x1 * x2 - y1 * y2 - z1 * z2
    x = w1 * x2 + x1 * w2 + y1 * z2 - z1 * y2
    y = w1 * y2 - x1 * z2 + y1 * w2 + z1 * x2
    z = w1 * z2 + x1 * y2 - y1 * x2 + z1 * w2
    return np.array([w, x, y, z])

@jit
def quaternion_conjugate(q):
    """
    Calculate the conjugate of a quaternion.

    Args:
        q (np.ndarray): The quaternion.

    Returns:
        np.ndarray: The conjugate of the quaternion.
    """
    return np.array([q[0], -q[1], -q[2], -q[3]])

@jit
def quaternion_between(v1, v2):
    """
    Calculate the quaternion representing the rotation from vector v1 to vector v2.

    Args:
        v1 (np.ndarray): The initial vector.
        v2 (np.ndarray): The final vector.

    Returns:
        np.ndarray: The quaternion representing the rotation.
    """
    u1 = ax.unit_vector_jit(v1)
    u2 = ax.unit_vector_jit(v2)

    if np.allclose(u1, u2):
        return np.array([1.0, 0.0, 0.0, 0.0])

    axis = np.cross(u1, u2)
    if np.linalg.norm(axis) == 0:
        axis = np.array([1.0, 0.0, 0.0])
    angle = axana.angle_between(u1, u2)

    w = np.cos(angle / 2)
    v = np.sin(angle / 2) * axis
    return np.array([w, *v])

@jit
def apply_quaternion(v, q):
    """
    Apply a quaternion rotation to a vector.

    Args:
        v (np.ndarray): The vector to rotate.
        q (np.ndarray): The quaternion representing the rotation.

    Returns:
        np.ndarray: The rotated vector.
    """
    if np.allclose(q, np.array([1.0, 0.0, 0.0, 0.0], dtype='float64')):
        return v
    
    v_q = np.array([0.0, *v])
    
    q_conj = quaternion_conjugate(q)
    q_inv = q_conj / np.linalg.norm(q_conj)**2
    rotated = quaternion_multiply(quaternion_multiply(q, v_q), q_inv)
    return rotated[1:]

@jit
def clip_quaternion(q, radians):
    """
    Clip a quaternion to a maximum rotation angle in radians.

    Args:
        q (np.ndarray): The quaternion to clip.
        radians (float): The maximum rotation angle in radians.

    Returns:
        np.ndarray: The clipped quaternion.
    """
    angle = 2 * np.arccos(q[0])
    max_angle = radians

    if angle > max_angle:
        angle = max_angle

    axis = q[1:]
    axis = axis / np.linalg.norm(axis)

    w = np.cos(angle / 2)
    v = np.sin(angle / 2) * axis

    return np.array([w, *v])

@jit
def clip_quaternion_degrees(q, degrees):
    """
    Clip a quaternion to a maximum rotation angle in degrees.

    Args:
        q (np.ndarray): The quaternion to clip.
        degrees (float): The maximum rotation angle in degrees.

    Returns:
        np.ndarray: The clipped quaternion.
    """
    return clip_quaternion(q, np.radians(degrees))


import axinite as ax
import axinite.analysis as axana
import numpy as np
from numba import jit

@jit
def _kinetic_energy_at(n: np.float64, bodies: np.ndarray):
    """
    Calculate the kinetic energy at a specific time step for an array of bodies.

    Args:
        n (np.float64): The time step index.
        bodies (np.ndarray): Array of body dictionaries with mass and velocity.

    Returns:
        float: The kinetic energy at the specified time step.
    """
    kinetic_energy = 0
    for body in bodies:
        kinetic_energy += 0.5 * body["m"] * np.linalg.norm(body["v"][n]) ** 2
    return kinetic_energy

def kinetic_energy_at(n: np.float64, bodies: list[ax.Body]):
    """
    Calculate the kinetic energy at a specific time step for a list of bodies.

    Args:
        n (np.float64): The time step index.
        bodies (list[ax.Body]): List of Body objects.

    Returns:
        float: The kinetic energy at the specified time step.
    """
    return _kinetic_energy_at(n, ax.get_inner_bodies(bodies))

@jit
def _total_kinetic_energy_at(n: np.float64, bodies: np.ndarray):
    """
    Calculate the total kinetic energy at a specific time step for an array of bodies.

    Args:
        n (np.float64): The time step index.
        bodies (np.ndarray): Array of body dictionaries with mass and velocity.

    Returns:
        float: The total kinetic energy at the specified time step.
    """
    return np.sum(_kinetic_energy_at(n, bodies))

def total_kinetic_energy_at(n: np.float64, bodies: list[ax.Body]):
    """
    Calculate the total kinetic energy at a specific time step for a list of bodies.

    Args:
        n (np.float64): The time step index.
        bodies (list[ax.Body]): List of Body objects.

    Returns:
        float: The total kinetic energy at the specified time step.
    """
    return _total_kinetic_energy_at(n, ax.get_inner_bodies(bodies))

@jit
def _kinetic_energy(bodies: np.ndarray):
    """
    Calculate the kinetic energy over all time steps for an array of bodies.

    Args:
        bodies (np.ndarray): Array of body dictionaries with mass and velocity.

    Returns:
        np.ndarray: Array of kinetic energies for each time step.
    """
    kinetic_energies = np.zeros(bodies[0]["r"].shape[0])
    n = 0
    while n < bodies[0]["r"].shape[0]:
        kinetic_energies[n] = _kinetic_energy_at(n, bodies)
        n += 1
    return kinetic_energies

def kinetic_energy(bodies: list[ax.Body]):
    """
    Calculate the kinetic energy over all time steps for a list of bodies.

    Args:
        bodies (list[ax.Body]): List of Body objects.

    Returns:
        np.ndarray: Array of kinetic energies for each time step.
    """
    return _kinetic_energy(ax.get_inner_bodies(bodies))

@jit
def _total_kinetic_energy(bodies: np.ndarray):
    """
    Calculate the total kinetic energy over all time steps for an array of bodies.

    Args:
        bodies (np.ndarray): Array of body dictionaries with mass and velocity.

    Returns:
        float: The total kinetic energy over all time steps.
    """
    return np.sum(_kinetic_energy(bodies))

def total_kinetic_energy(bodies: list[ax.Body]):
    """
    Calculate the total kinetic energy over all time steps for a list of bodies.

    Args:
        bodies (list[ax.Body]): List of Body objects.

    Returns:
        float: The total kinetic energy over all time steps.
    """
    return _total_kinetic_energy(ax.get_inner_bodies(bodies))

@jit
def _potential_energy_at(n: np.float64, bodies: np.ndarray):
    """
    Calculate the potential energy at a specific time step for an array of bodies.

    Args:
        n (np.float64): The time step index.
        bodies (np.ndarray): Array of body dictionaries with mass and position.

    Returns:
        float: The potential energy at the specified time step.
    """
    potential_energy = 0
    for i in range(len(bodies)):
        for j in range(i + 1, len(bodies)):
            distance = np.linalg.norm(bodies[i]["r"][n] - bodies[j]["r"][n])
            if distance > 0: 
                potential_energy -= ax.G * bodies[i]["m"] * bodies[j]["m"] / distance
    return potential_energy

def potential_energy_at(n: np.float64, bodies: list[ax.Body]):
    """
    Calculate the potential energy at a specific time step for a list of bodies.

    Args:
        n (np.float64): The time step index.
        bodies (list[ax.Body]): List of Body objects.

    Returns:
        float: The potential energy at the specified time step.
    """
    return _potential_energy_at(n, ax.get_inner_bodies(bodies))

@jit
def _total_potential_energy_at(n: np.float64, bodies: np.ndarray):
    """
    Calculate the total potential energy at a specific time step for an array of bodies.

    Args:
        n (np.float64): The time step index.
        bodies (np.ndarray): Array of body dictionaries with mass and position.

    Returns:
        float: The total potential energy at the specified time step.
    """
    return np.sum(_potential_energy_at(n, bodies))

def total_potential_energy_at(n: np.float64, bodies: list[ax.Body]):
    """
    Calculate the total potential energy at a specific time step for a list of bodies.

    Args:
        n (np.float64): The time step index.
        bodies (list[ax.Body]): List of Body objects.

    Returns:
        float: The total potential energy at the specified time step.
    """
    return _total_potential_energy_at(n, ax.get_inner_bodies(bodies))

@jit
def _potential_energy(bodies: np.ndarray):
    """
    Calculate the potential energy over all time steps for an array of bodies.

    Args:
        bodies (np.ndarray): Array of body dictionaries with mass and position.

    Returns:
        np.ndarray: Array of potential energies for each time step.
    """
    potential_energies = np.zeros(bodies[0]["r"].shape[0])
    n = 0
    while n < bodies[0]["r"].shape[0]:
        potential_energies[n] = _potential_energy_at(n, bodies)
        n += 1
    return potential_energies

def potential_energy(bodies: list[ax.Body]):
    """
    Calculate the potential energy over all time steps for a list of bodies.

    Args:
        bodies (list[ax.Body]): List of Body objects.

    Returns:
        np.ndarray: Array of potential energies for each time step.
    """
    return _potential_energy(ax.get_inner_bodies(bodies))

@jit
def _total_potential_energy(bodies: np.ndarray):
    """
    Calculate the total potential energy over all time steps for an array of bodies.

    Args:
        bodies (np.ndarray): Array of body dictionaries with mass and position.

    Returns:
        float: The total potential energy over all time steps.
    """
    return np.sum(_potential_energy(bodies))

def total_potential_energy(bodies: list[ax.Body]):
    """
    Calculate the total potential energy over all time steps for a list of bodies.

    Args:
        bodies (list[ax.Body]): List of Body objects.

    Returns:
        float: The total potential energy over all time steps.
    """
    return _total_potential_energy(ax.get_inner_bodies(bodies))

@jit
def _energy_at(n: np.float64, bodies: np.ndarray):
    """
    Calculate the total energy (kinetic + potential) at a specific time step for an array of bodies.

    Args:
        n (np.float64): The time step index.
        bodies (np.ndarray): Array of body dictionaries with mass, velocity, and position.

    Returns:
        float: The total energy at the specified time step.
    """
    return _kinetic_energy_at(n, bodies) + _potential_energy_at(n, bodies)

def energy_at(n: np.float64, bodies: list[ax.Body]):
    """
    Calculate the total energy (kinetic + potential) at a specific time step for a list of bodies.

    Args:
        n (np.float64): The time step index.
        bodies (list[ax.Body]): List of Body objects.

    Returns:
        float: The total energy at the specified time step.
    """
    return _energy_at(n, ax.get_inner_bodies(bodies))

@jit
def _total_energy_at(n: np.float64, bodies: np.ndarray):
    """
    Calculate the total energy (total kinetic + total potential) at a specific time step for an array of bodies.

    Args:
        n (np.float64): The time step index.
        bodies (np.ndarray): Array of body dictionaries with mass, velocity, and position.

    Returns:
        float: The total energy at the specified time step.
    """
    return _total_kinetic_energy_at(n, bodies) + _total_potential_energy_at(n, bodies)

def total_energy_at(n: np.float64, bodies: list[ax.Body]):
    """
    Calculate the total energy (total kinetic + total potential) at a specific time step for a list of bodies.

    Args:
        n (np.float64): The time step index.
        bodies (list[ax.Body]): List of Body objects.

    Returns:
        float: The total energy at the specified time step.
    """
    return _total_energy_at(n, ax.get_inner_bodies(bodies))

@jit
def _energy(bodies: np.ndarray):
    """
    Calculate the total energy (kinetic + potential) over all time steps for an array of bodies.

    Args:
        bodies (np.ndarray): Array of body dictionaries with mass, velocity, and position.

    Returns:
        np.ndarray: Array of total energies for each time step.
    """
    return _kinetic_energy(bodies) + _potential_energy(bodies)

def energy(bodies: list[ax.Body]):
    """
    Calculate the total energy (kinetic + potential) over all time steps for a list of bodies.

    Args:
        bodies (list[ax.Body]): List of Body objects.

    Returns:
        np.ndarray: Array of total energies for each time step.
    """
    return _energy(ax.get_inner_bodies(bodies))

@jit
def _total_energy(bodies: np.ndarray):
    """
    Calculate the total energy (total kinetic + total potential) over all time steps for an array of bodies.

    Args:
        bodies (np.ndarray): Array of body dictionaries with mass, velocity, and position.

    Returns:
        float: The total energy over all time steps.
    """
    return _total_kinetic_energy(bodies) + _total_potential_energy(bodies)

def total_energy(bodies: list[ax.Body]):
    """
    Calculate the total energy (total kinetic + total potential) over all time steps for a list of bodies.

    Args:
        bodies (list[ax.Body]): List of Body objects.

    Returns:
        float: The total energy over all time steps.
    """
    return _total_energy(ax.get_inner_bodies(bodies))

def kinetic_energy_of(body: ax.Body):
    """
    Calculate the kinetic energy over all time steps for a single body.

    Args:
        body (ax.Body): A Body object.

    Returns:
        np.ndarray: Array of kinetic energies for each time step.
    """
    inner_body = ax.get_inner_body(body)
    kinetic_energies = np.zeros(inner_body["r"].shape[0])
    for n in range(inner_body["r"].shape[0]):
        kinetic_energies[n] = 0.5 * inner_body["m"] * np.linalg.norm(inner_body["v"][n]) ** 2
    return kinetic_energies

def potential_energy_of(body: ax.Body, other_bodies: list[ax.Body]):
    """
    Calculate the potential energy over all time steps for a single body with respect to other bodies.

    Args:
        body (ax.Body): A Body object.
        other_bodies (list[ax.Body]): List of other Body objects.

    Returns:
        np.ndarray: Array of potential energies for each time step.
    """
    inner_body = ax.get_inner_body(body)
    inner_other_bodies = ax.get_inner_bodies(other_bodies)
    potential_energies = np.zeros(inner_body["r"].shape[0])
    for n in range(inner_body["r"].shape[0]):
        potential_energy = 0
        for other_body in inner_other_bodies:
            distance = np.linalg.norm(inner_body["r"][n] - other_body["r"][n])
            if distance > 0:
                potential_energy -= ax.G * inner_body["m"] * other_body["m"] / distance
        potential_energies[n] = potential_energy
    return potential_energies

def total_energy_of(body: ax.Body, other_bodies: list[ax.Body]):
    """
    Calculate the total energy (kinetic + potential) over all time steps for a single body with respect to other bodies.

    Args:
        body (ax.Body): A Body object.
        other_bodies (list[ax.Body]): List of other Body objects.

    Returns:
        np.ndarray: Array of total energies for each time step.
    """
    kinetic_energies = kinetic_energy_of(body)
    potential_energies = potential_energy_of(body, other_bodies)
    return kinetic_energies + potential_energies


import axinite as ax
import axinite.tools as axtools
import json

def read(path: str) -> axtools.AxiniteArgs:
    """
    Read a simulation from a file.

    Args:
        path (str): The path to read from.

    Returns:
        AxiniteArgs: The simulation result.
    """
    try:
        with open(path, "r") as file:
            return reads(file.read())
    except FileNotFoundError as e:
        print(f"Error: File {path} not found.")
        raise e

def reads(string: str) -> axtools.AxiniteArgs:
    """
    Read a simulation from a string.

    Args:
        string (str): The string to read from.

    Returns:
        AxiniteArgs: The simulation result.
    """
    
    try: data = json.loads(string)
    except json.JSONDecodeError as e: 
        print("Error: Invalid JSON string.") 
        raise e
    
    args = axtools.AxiniteArgs()
    try:
        args.name = data["name"]
        args.delta = ax.interpret_time(data["delta"])
        args.limit = ax.round_limit(ax.interpret_time(data["limit"]), args.delta)
        args.t = data["t"]

        if "radius_multiplier" in data:
            args.radius_multiplier = data["radius_multiplier"]

        if "rate" in data:
            args.rate = data["rate"]

        if "retain" in data:
            args.retain = data["retain"]

        if "frontend_args" in data:
            args.frontend_args = data["frontend_args"]

        for body in data["bodies"]: 
            args.bodies.append(axtools.data_to_body(body, args.limit, args.delta))
    except KeyError as e:
        print(f"Error: {e} not found in the provided string.")
        raise e

    return args


import axinite.tools as axtools
import axinite as ax
import numpy as np

class AxiniteArgs:
    """
    A class to store simulation parameters for the Axinite celestial mechanics engine.

    Attributes:
        name (str): The name of the simulation.
        delta (np.float64): The frequency at which the simulation should be computed in seconds.
        limit (np.float64): The length of the simulation in seconds.
        action (function): A function to be called at each timestep.
        t (np.float64): The current timestep.
        bodies (list[axtools.Body]): A list of Body objects to be simulated.
        radius_multiplier (float): A float to multiply all radii by.
        rate (int): The number of frames per second to render for the live and run functions.
        retain (int): The number of points to retain on each body's trail.
        modifier (function): A function called to modify the forces on the bodies.
        frontend_args (dict[str, dict[str, str|float|int|bool|list|dict]]): A dictionary of frontend-specific arguments.
        backend (function): The backend (integration method) to use.
        action_frequency (int): The frequency at which the action function should be called.
    """
    
    def __init__(self):
        """Initializes an AxiniteArgs object."""

        self.name: str = None
        "The name of the simulation."

        self.delta: np.float64 = None
        "The frequency at which the simulation should be computed in seconds."

        self.limit: np.float64 = None
        "The length of the simulation in seconds."

        self.action: function = None
        "A function to be called at each timestep."

        self.t: np.float64 = None
        "The current timestep."

        self.bodies: list[axtools.Body] = []
        "A list of Body objects to be simulated."

        self.radius_multiplier: float = None
        "A float to multiply all radii by."

        self.rate: int = None
        "The number frames per second to render for the live and run functions."

        self.retain: int = None
        "The number of points to retain on each body's trail."

        self.modifier: function = None
        "A function called to modify the forces on the bodies."

        self.frontend_args: dict[str, dict[str, str|float|int|bool|list|dict]] = {}
        "A dictionary of frontend-specific arguments."

        self.backend: function = ax.verlet_backend
        "The backend (integration method) to use."

        self.action_frequency: int = None
        "The frequency at which the action function should be called."
    
    def set_limit(self, limit: np.float64) -> None:
        """
        Sets the limit of the simulation.

        Args:
            limit (np.float64): The length of the simulation in seconds.
        """
        _bodies = []
        self.limit = limit
        for body in self.bodies:
            body = axtools.Body(body.name, body.mass, limit, self.delta, body._inner["r"][0], body._inner["v"][0])
            _bodies.append(body)
        self.bodies = _bodies
    
    def set_delta(self, delta: np.float64) -> None:
        """
        Sets the delta of the simulation.

        Args:
            delta (np.float64): The frequency at which the simulation should be computed in seconds.
        """
        _bodies = []
        self.delta = delta
        self.set_limit(ax.round_limit(self.limit, delta))
        for body in self.bodies:
            body = axtools.Body(body.name, body.mass, self.limit, delta, body._inner["r"][0], body._inner["v"][0])
            _bodies.append(body)
        self.bodies = _bodies

    def unpack(self) -> tuple[np.float64, np.float64, 'function', '*tuple[axtools.Body, ...]']:
        """
        Unpacks the AxiniteArgs object into a tuple that can be passed to `axinite`'s load function.

        Returns:
            tuple: A tuple containing delta, limit, action, and bodies.
        """
        return self.delta, self.limit, self.backend, *self.bodies


"""
The `axinite` module provides the core functionality for the Axinite celestial mechanics engine.

This module includes classes and functions for representing celestial bodies, performing numerical integration
using various methods, and loading simulation data.
"""

__version__ = "1.24.0"

from axinite.body import Body
from axinite.functions import vector_magnitude_jit, unit_vector_jit, gravitational_force_jit, body_dtype, \
    get_inner_bodies, _body, create_outer_bodies, timestep, interpret_distance, interpret_mass, interpret_time, \
    timesteps, clip_scalar, G, state, time_to, mass_to, distance_to, round_limit, gravitational_forces, locate_body
import axinite.functions as functions
from axinite.load import load
from axinite.backends.euler import euler_backend, euler_nojit_backend
from axinite.backends.verlet import verlet_backend, verlet_nojit_backend
from axinite.backends.rk2 import rk2_backend, rk2_nojit_backend
from axinite.backends.rk3 import rk3_backend, rk3_nojit_backend
from axinite.backends.rk4 import rk4_backend, rk4_nojit_backend
import axinite.backends as backends
import axinite.analysis as analysis
import axinite.tools as tools
import axinite.utils as utils
import axinite.analysis as axana
import axinite.tools as axtools
import axinite.utils as axutils
import axinite as ax


import axinite as ax
import numpy as np
from numba import jit

@jit
def angle_between(v1, v2):
    """Calculates the angle between two vectors in radians.

    Args:
        v1 (np.ndarray): The first vector.
        v2 (np.ndarray): The second vector.

    Returns:
        float: The angle between the two vectors in radians.
    """
    unit_v1 = ax.unit_vector_jit(v1)
    unit_v2 = ax.unit_vector_jit(v2)
    dot_product = np.dot(unit_v1, unit_v2)
    angle_rad = np.arccos(ax.clip_scalar(dot_product, -1.0, 1.0))
    return angle_rad

@jit
def angle_between_degrees(v1, v2):
    """Calculates the angle between two vectors in degrees.

    Args:
        v1 (np.ndarray): The first vector.
        v2 (np.ndarray): The second vector.

    Returns:
        float: The angle between the two vectors in degrees.
    """
    angle_rad = angle_between(v1, v2)
    angle_deg = np.degrees(angle_rad)
    return angle_deg